#!/usr/bin/env python3

import json
from pathlib import Path
from .witlogger import getLogger

log = getLogger()


# TODO
# Should this actually be shared between package manifests and workspace descriptions?
# Should we use different datastructures?
class Manifest:
    """
    Common class for the description of package dependencies and a workspace
    """

    def __init__(self, dependencies):
        self.dependencies = dependencies

    def get_dependency(self, name: str):
        for d in self.dependencies:
            if d.name == name:
                return d
        return None

    def contains_dependency(self, name: str) -> bool:
        return self.get_dependency(name) is not None

    def add_dependency(self, dep):
        resolved = dep.resolved()
        log.debug("Adding to manifest: {}".format(resolved))
        self.dependencies.append(resolved)

    def replace_dependency(self, dep) -> None:
        newdeps = []
        found = False
        for d in self.dependencies:
            if d.name == dep.name:
                resolved = dep.resolved()
                log.debug("New replace dep: {}".format(resolved))
                newdeps.append(resolved)
                found = True
            else:
                newdeps.append(d)
        assert found, \
            "Trying to update '{}' but it doesn't exist in manifest!".format(dep.name)
        self.dependencies = newdeps

    def write(self, path):
        contents = [p.manifest() for p in self.dependencies]
        manifest_json = json.dumps(contents, sort_keys=True, indent=4) + '\n'
        path.write_text(manifest_json)

    @staticmethod
    def read_manifest(path, safe=False):
        if safe and not Path(path).exists():
            return Manifest([])
        content = json.loads(path.read_text())
        return Manifest.process_manifest(content, path)

    @staticmethod
    def process_manifest(json_content, location):
        # Fail if there are dependencies with the same name
        dep_names = [dep['name'] for dep in json_content]
        if len(dep_names) != len(set(dep_names)):
            dup = set([x for x in dep_names if dep_names.count(x) > 1])
            raise Exception("Two dependencies have the same name in '{}': {}".format(location, dup))

        # import here to prevent circular dependency
        from .dependency import manifest_item_to_dep
        dependencies = [manifest_item_to_dep(x) for x in json_content]
        return Manifest(dependencies)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python3

import subprocess
from pathlib import Path
from pprint import pformat
import json
import re
from . import manifest
from .common import WitUserError
from .witlogger import getLogger
from typing import Set  # noqa: F401
from functools import lru_cache
from .env import git_reference_workspace

log = getLogger()


class GitError(Exception):
    pass


class GitCommitNotFound(Exception):
    pass


class BadSource(WitUserError):
    def __init__(self, name, source):
        self.name = name
        self.source = source

    def __str__(self):
        return "Bad remote for '{}':\n  {}".format(self.name, self.source)


verbose_prefix = re.compile(r"^refs/(?:heads/)?")


# TODO Could speed up validation
#   - use git ls-remote to validate remote exists
#   - use git ls-remote to validate revision for tags and branches
#   - if github repo, check if page exists (or if you get 404)

class GitRepo:
    """
    In memory data structure representing a Git repo package
    It may not be in sync with data structures on the file system
    Note there can be multiple GitRepo objects for the same package
    """
    PKG_DEPENDENCY_FILE = "wit-manifest.json"

    def __init__(self, name, wsroot: Path):
        self.name = name
        self.path = wsroot / name
        # Cache known hashes for quick lookup
        self._known_hashes = set()  # type: Set[str]

    def _known_hash(self, commit) -> bool:
        """Checks if a hash exists in the current repo"""
        return commit in self._known_hashes

    def _add_known_hash(self, commit):
        self._known_hashes.add(commit)

    def is_bad_source(self, source):
        tmp = self.path
        self.path = self.path.parent
        proc = self._git_command('ls-remote', source)
        self.path = tmp
        return proc.returncode != 0

    # name is needed for generating error messages
    def download(self, source, name):
        if not GitRepo.is_git_repo(self.path):
            self.clone(source, name)
        self.fetch(source, name)

    # name is needed for generating error messages
    def clone(self, source, name):
        assert not GitRepo.is_git_repo(self.path), \
            "Trying to clone and checkout into existing git repo!"

        cmd = ["clone", *self._git_reference_options(), "--no-checkout", source, str(self.path)]
        proc = self._git_command(*cmd, working_dir=str(self.path.parent))
        try:
            self._git_check(proc)
        except GitError:
            if self.is_bad_source(source):
                raise BadSource(name, source)
            else:
                raise
        log.info('Cloned {}'.format(self.name))

    def _git_reference_options(self):
        """
        Use git clone's '--reference' to point at a local repository cache to copy objects/commits
        to save network traffic. Any missing objects/commits are downloaded from the true remote.
        Only newer git versions can use '--reference-if-able', so we emulate the 'if-able' bit.
        """
        if not git_reference_workspace:
            return []
        paths = [Path(git_reference_workspace) / self.name,
                 Path(git_reference_workspace) / (self.name+'.git')]
        for path in paths:
            if path.is_dir():
                return ["--reference", str(path), "--dissociate"]
        return []

    # name is needed for generating error messages
    def fetch(self, source, name):
        # in case source is a remote and we want a commit
        proc = self._git_command('fetch', source)
        # in case source is a file path and we want, for example, origin/master
        self._git_command('fetch', '--all')
        try:
            self._git_check(proc)
        except GitError:
            if self.is_bad_source(source):
                raise BadSource(name, source)
            else:
                raise
        return proc.returncode == 0

    def get_head_commit(self) -> str:
        return self.get_commit('HEAD')

    @lru_cache(maxsize=None)
    def _get_commit_cached(self, commit):
        return self._get_commit_impl(commit)

    def _get_commit_impl(self, commit):
        proc = self._git_command('rev-parse', commit)
        try:
            self._git_check(proc)
        except GitError:
            proc = self._git_command('rev-parse', 'origin/{}'.format(commit))
            try:
                self._git_check(proc)
            except GitError:
                if 'unknown revision or path not in the working tree' in proc.stderr:
                    raise GitCommitNotFound
                else:
                    raise
        return proc.stdout.rstrip()

    def get_commit(self, commit) -> str:
        if self._known_hash(commit):
            result = self._get_commit_cached(commit)
        else:
            result = self._get_commit_impl(commit)
        self._add_known_hash(result)
        return result

    @lru_cache(maxsize=None)
    def _get_shortened_rev_cached(self, commit):
        return self._get_shortened_rev_impl(commit)

    def _get_shortened_rev_impl(self, commit):
        proc = self._git_command('rev-parse', '--short', commit)
        self._git_check(proc)
        return proc.stdout.rstrip()

    def get_shortened_rev(self, commit):
        if self._known_hash(commit):
            return self._get_shortened_rev_cached(commit)
        else:
            return self._get_shortened_rev_impl(commit)

    def is_hash(self, ref):
        return self.get_commit(ref) == ref

    def is_tag(self, ref):
        proc = self._git_command('tag', '--list', ref)
        self._git_check(proc)
        return ref in proc.stdout.split('\n')

    def has_commit(self, commit) -> bool:
        # rev-parse does not always fail when a commit is missing
        proc = self._git_command('cat-file', '-t', commit)
        return proc.returncode == 0

    def have_common_ancestor(self, commits):
        proc = self._git_command('merge-base', '--octopus', *commits)
        return proc.returncode == 0

    def get_remote(self) -> str:
        # TODO Do we need to worry about other remotes?
        proc = self._git_command('remote', 'get-url', 'origin')
        self._git_check(proc)
        return proc.stdout.rstrip()

    def set_origin(self, source):
        proc = self._git_command('remote', 'set-url', 'origin', source)
        self._git_check(proc)

    def clean(self):
        proc = self._git_command('status', '--porcelain')
        self._git_check(proc)
        return proc.stdout == ""

    def modified(self):
        proc = self._git_command('status', '--porcelain')
        self._git_check(proc)
        for line in proc.stdout.split("\n"):
            if line.lstrip().startswith("M"):
                return True
        return False

    def untracked(self):
        proc = self._git_command('status', '--porcelain')
        self._git_check(proc)
        for line in proc.stdout.split("\n"):
            if line.lstrip().startswith("??"):
                return True
        return False

    def modified_manifest(self):
        proc = self._git_command('status', '--porcelain')
        self._git_check(proc)
        for line in proc.stdout.split("\n"):
            if ((line.lstrip().startswith("M") or line.lstrip().startswith("D"))
                    and line.endswith("wit-manifest.json")):
                return True
        return False

    @lru_cache(maxsize=None)
    def _commit_to_time_cached(self, hash):
        return self._commit_to_time_impl(hash)

    def _commit_to_time_impl(self, hash):
        proc = self._git_command('log', '-n1', '--format=%ct', hash)
        self._git_check(proc)
        return proc.stdout.rstrip()

    def commit_to_time(self, hash):
        if self._known_hash(hash):
            return self._commit_to_time_cached(hash)
        else:
            return self._commit_to_time_impl(hash)

    def is_ancestor(self, ancestor, current=None):
        proc = self._git_command("merge-base", "--is-ancestor", ancestor,
                                 current or self.get_head_commit())
        return proc.returncode == 0

    def read_manifest(self) -> manifest.Manifest:
        mpath = self.manifest_path()
        return manifest.Manifest.read_manifest(mpath, safe=True)

    def write_manifest(self, manifest) -> None:
        mpath = self.manifest_path()
        manifest.write(mpath)

    def read_manifest_from_commit(self, revision) -> manifest.Manifest:
        proc = self._git_command("show", "{}:{}".format(revision, GitRepo.PKG_DEPENDENCY_FILE))
        if proc.returncode:
            log.debug("No dependency file found in repo [{}:{}]".format(revision,
                      self.path))
        json_content = [] if proc.returncode else json.loads(proc.stdout)
        return manifest.Manifest.process_manifest(json_content, self.name)

    def checkout(self, revision):
        wanted_hash = self.get_commit(revision)
        if self.get_commit('HEAD') != wanted_hash:
            proc_ref = self._git_command("show-ref")
            self._git_check(proc_ref)
            rev_names = proc_ref.stdout.rstrip().split('\n')
            rev_names = [r.split(' ') for r in rev_names]
            rev_names = [r[1] for r in rev_names if r[0] == wanted_hash]
            rev_names = [r for r in rev_names if not r.startswith('refs/remotes')]
            rev_names = [verbose_prefix.sub('', r) for r in rev_names]

            suggestions = ''
            if len(rev_names) > 1:
                suggestions = ' ({})'.format(', '.join(rev_names))

            if len(rev_names) != 1:
                rev = revision
                log.info("Checking out '{}' at '{}'{}".format(self.name, rev, suggestions))
            else:
                rev = rev_names[0]
                log.info("Checking out '{}' at '{}' ({})".format(self.name, rev, revision))

            proc = self._git_command("checkout", rev)
            self._git_check(proc)
        else:
            proc = self._git_command("checkout")
            self._git_check(proc)

        # If our revision was a branch or tag, get the actual commit
        self.revision = self.get_head_commit()

    def manifest_path(self):
        return self.path / self.PKG_DEPENDENCY_FILE

    def manifest(self, source, revision):
        return {
            'name': self.name,
            'source': source,
            'commit': revision,
        }

    def _git_command(self, *args, working_dir=None):
        cwd = str(self.path) if working_dir is None else str(working_dir)
        log.debug("Executing [{}] in [{}]".format(' '.join(['git', *args]), cwd))
        proc = subprocess.run(['git', *args],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              universal_newlines=True,
                              cwd=cwd)
        log.spam("   stderr: [{}]".format(proc.stderr.rstrip()))
        log.spam("   stdout: [{}]".format(proc.stdout.rstrip()))
        return proc

    def _git_check(self, proc):
        if proc.returncode:
            msg = "Command [{}] in [{}] exited with non-zero exit status [{}]\n".format(
                  ' '.join(proc.args), str(self.path), proc.returncode)
            msg += "stdout: [{}]\n".format(proc.stdout.rstrip())
            msg += "stderr: [{}]\n".format(proc.stderr.rstrip())
            raise GitError(msg)

        return proc.returncode

    @staticmethod
    def path_to_name(path):
        """
        >>> GitRepo.path_to_name("a.git")
        'a'
        >>> GitRepo.path_to_name("/a/b/c/def.git")
        'def'
        >>> GitRepo.path_to_name("ghi")
        'ghi'
        """
        return Path(path).name.replace('.git', '')

    @staticmethod
    def is_git_repo(path):
        cmd = ['git', 'ls-remote', '--exit-code', str(path)]
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        ret = proc.returncode
        return ret == 0

    # Enable prettyish-printing of the class
    def __repr__(self):
        return pformat(vars(self), indent=4, width=1)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python3

# TODO:
# * Error handling
# * Log all commands into workspace
# * De-couple WorkSpace from GitRepos
# * Write unit tests
# * Handle corrupt dependencies.json
# * Handle exceptional conditions
# * Use a real logger
# * Handle partial sha1s correctly

import subprocess
import sys
import shutil
import os
from .witlogger import getLogger
from .workspace import WorkSpace, PackageNotInWorkspaceError
from .dependency import Dependency
from .inspect import inspect_tree
from pathlib import Path
from typing import cast, List, Tuple  # noqa: F401
from .common import error, WitUserError, print_errors
from .gitrepo import GitRepo, GitCommitNotFound
from .manifest import Manifest
from .package import WitBug
from .parser import parser, add_dep_parser
from .version import __version__
import re

log = getLogger()


class NotAPackageError(WitUserError):
    pass


def main() -> None:

    args = parser.parse_args()
    if args.verbose >= 4:
        log.setLevel('SPAM')
    elif args.verbose == 3:
        log.setLevel('TRACE')
    elif args.verbose == 2:
        log.setLevel('DEBUG')
    elif args.verbose == 1:
        log.setLevel('VERBOSE')
    else:
        log.setLevel('INFO')

    log.debug("Log level: {}".format(log.getLevelName()))

    if args.prepend_repo_path and args.repo_path:
        args.repo_path = " ".join([args.prepend_repo_path, args.repo_path])
    elif args.prepend_repo_path:
        args.repo_path = args.prepend_repo_path

    if args.version:
        version()
        sys.exit(0)

    try:
        # FIXME: This big switch statement... no good.
        if args.command == 'init':
            create(args)
        elif args.command == 'restore':
            restore_from_lock(args)

        else:
            # These commands assume the workspace already exists. Error out if the
            # workspace cannot be found.
            try:
                ws = WorkSpace.find(Path.cwd(), parse_repo_path(args), args.jobs)

            except FileNotFoundError as e:
                log.error("Unable to find workspace root [{}]. Cannot continue.".format(e))
                sys.exit(1)

            if args.command == 'add-pkg':
                add_pkg(ws, args)

            elif args.command == 'update-pkg':
                update_pkg(ws, args)

            elif args.command == 'add-dep':
                add_dep(ws, args)

            elif args.command == 'update-dep':
                update_dep(ws, args)

            elif args.command == 'status':
                status(ws, args)

            elif args.command == 'update':
                update(ws, args)

            elif args.command == 'foreach':
                foreach(ws, args)

            elif args.command == 'inspect':
                if args.dot or args.tree:
                    inspect_tree(ws, args)
                else:
                    log.error('`wit inspect` must be run with a flag')
                    print(parser.parse_args('inspect -h'.split()))
                    sys.exit(1)
    except WitUserError as e:
        error(e)
    except AssertionError as e:
        raise WitBug(e)


def foreach(ws, args):
    has_fail = False
    for pkg in ws.lock.packages:
        env = os.environ.copy()
        env["WIT_REPO_NAME"] = pkg.name
        env["WIT_REPO_PATH"] = str(ws.root / pkg.name)
        env["WIT_LOCK_SOURCE"] = pkg.source
        env["WIT_LOCK_COMMIT"] = pkg.revision
        env["WIT_WORKSPACE"] = str(ws.root)

        log.info("Entering '{}'".format(pkg.name))

        location = str(ws.root / pkg.name)
        proc = subprocess.run([args.cmd, *args.args], shell=True,
                              env=env, cwd=location, universal_newlines=True)

        if proc.returncode != 0:
            has_fail = True
            command = " ".join([args.cmd, *args.args])
            log.error("Command '{}' in '{}' failed with exitcode: {}"
                      .format(command, location, proc.returncode))
            if not args.continue_on_fail:
                sys.exit(proc.returncode)

    if has_fail:
        sys.exit(1)


def parse_repo_path(args):
    return args.repo_path.split(' ') if args.repo_path else []


def create(args) -> None:
    if args.add_pkg is None:
        dependencies = []  # type: List[Tuple[str, str]]
    else:
        dependencies = args.add_pkg

    ws = WorkSpace.create(args.workspace_name, parse_repo_path(args), args.jobs)
    for dep in dependencies:
        ws.add_dependency(dep)

    if args.update:
        update(ws, args)


# A user can restore a workspace in the current directory, or in a new directory.
# A wit-lock.json and wit-workspace.json needs to be found either in the current directory
# or separately specified by arguments.
def restore_from_lock(args) -> None:
    current_dir = Path.cwd()
    lock_dir = current_dir
    dest_ws = current_dir

    if args.workspace_name:
        dest_ws = current_dir / args.workspace_name
        if dest_ws.exists():
            log.error("New workspace directory [{}] already exists.".format(str(dest_ws)))
            sys.exit(1)
        else:
            log.info("Creating new workspace [{}]".format(str(dest_ws)))
            dest_ws.mkdir()

    dotwit = dest_ws/'.wit'
    if dotwit.exists():
        log.error("Directory [{}] is already a workspace, contains a .wit directory."
                  .format(str(dest_ws)))
        sys.exit(1)
    dotwit.mkdir()

    if args.from_workspace:
        lock_dir = Path(args.from_workspace)

    ws = 'wit-workspace.json'
    lock = 'wit-lock.json'
    if not (lock_dir/lock).exists():
        log.error("Could not find {}".format(str(lock_dir/lock)))
        sys.exit(1)
    if not (lock_dir/ws).exists():
        log.error("Could not find {}".format(str(lock_dir/ws)))
        sys.exit(1)

    if args.from_workspace or args.workspace_name:
        shutil.copy(str(lock_dir/ws), str(dest_ws/ws))
        shutil.copy(str(lock_dir/lock), str(dest_ws/lock))

    WorkSpace.restore(dest_ws)


def add_pkg(ws, args) -> None:
    log.info("Adding package to workspace")
    ws.add_dependency(args.repo)


def update_pkg(ws, args) -> None:
    ws.update_dependency(args.repo)


def dependency_from_tag(wsroot, tag, message=None):
    source, revision = tag

    dotwit = wsroot / ".wit"
    if (wsroot/source).exists() and (wsroot/source).parent == wsroot:
        repo = GitRepo((wsroot/source).name, wsroot)
        source = repo.get_remote()
    elif (dotwit/source).exists() and (dotwit/source).parent == dotwit:
        repo = GitRepo((dotwit/source).name, dotwit)
        source = repo.get_remote()
    elif (wsroot/source).exists():
        source = str((wsroot/source).resolve())
    elif Path(source).exists():
        source = str(Path(source).resolve())

    return Dependency(None, source, revision, message)


def add_dep(ws, args) -> None:
    """ Resolve a Dependency then add it to the cwd's wit-manifest.json """
    packages = {pkg.name: pkg for pkg in ws.lock.packages}
    req_dep = dependency_from_tag(ws.root, args.pkg, message=args.message)

    cwd = Path(os.getcwd()).resolve()
    if cwd == ws.root:
        error("add-dep must be run inside of a package, not the workspace root.\n\n" +
              add_dep_parser.format_help())
    cwd_dirname = cwd.relative_to(ws.root).parts[0]
    if not ws.lock.contains_package(cwd_dirname):
        raise NotAPackageError(
            "'{}' is not a package in workspace at '{}'".format(cwd_dirname, ws.path))

    # in order to resolve the revision, we need to bind
    # the req_dep to disk, cloning into .wit if neccesary
    req_dep.load(packages, ws.repo_paths, ws.root, True)
    try:
        req_dep.package.revision = req_dep.resolved_rev()
    except GitCommitNotFound:
        raise WitUserError("Could not find commit or reference '{}' in '{}'"
                           "".format(req_dep.specified_revision, req_dep.name))

    manifest_path = cwd/'wit-manifest.json'
    if manifest_path.exists():
        manifest = Manifest.read_manifest(manifest_path)
    else:
        manifest = Manifest([])

    # make sure the dependency is not already in the cwd's manifest
    if manifest.contains_dependency(req_dep.name):
        log.error("'{}' already depends on '{}'".format(cwd_dirname, req_dep.name))
        sys.exit(1)

    manifest.add_dependency(req_dep)
    manifest.write(manifest_path)

    log.info("'{}' now depends on '{}'".format(cwd_dirname, req_dep.package.id()))


def update_dep(ws, args) -> None:
    packages = {pkg.name: pkg for pkg in ws.lock.packages}
    req_dep = dependency_from_tag(ws.root, args.pkg, message=args.message)

    cwd = Path(os.getcwd()).resolve()

    if cwd == ws.root:
        error("update-dep must be run inside of a package, not the workspace root.\n"
              "  A dependency is updated in the package determined by the current working "
              "directory,\n  which can also be set by -C.")

    cwd_dirname = cwd.relative_to(ws.root).parts[0]
    manifest = Manifest.read_manifest(cwd/'wit-manifest.json')

    # make sure the package is already in the cwd's manifest
    if not manifest.contains_dependency(req_dep.name):
        log.error("'{}' does not depend on '{}'".format(cwd_dirname, req_dep.name))
        sys.exit(1)

    req_dep.load(packages, ws.repo_paths, ws.root, True)
    req_pkg = req_dep.package
    try:
        req_pkg.revision = req_dep.resolved_rev()
    except GitCommitNotFound:
        raise WitUserError("Could not find commit or reference '{}' in '{}'"
                           "".format(req_dep.specified_revision, req_dep.name))

    # check if the requested repo is missing from disk
    if req_pkg.repo is None:
        msg = "'{}' not found in workspace. Have you run 'wit update'?".format(req_dep.name)
        raise PackageNotInWorkspaceError(msg)

    log.info("Updating to {}".format(req_dep.resolved_rev()))
    manifest.replace_dependency(req_dep)
    manifest.write(cwd/'wit-manifest.json')

    log.info("'{}' now depends on '{}'".format(cwd_dirname, req_pkg.id()))


def status(ws, args) -> None:
    log.debug("Checking workspace status")
    if not ws.lock:
        log.info("{} is empty. Have you run `wit update`?".format(ws.LOCK))
        return

    clean = []
    dirty = []
    untracked = []
    missing = []
    seen_paths = {}
    for package in ws.lock.packages:
        package.load(ws.root, False)
        if package.repo is None:
            missing.append(package)
            continue
        seen_paths[package.repo.path] = True

        lock_commit = package.revision
        latest_commit = package.repo.get_head_commit()

        new_commits = lock_commit != latest_commit

        if new_commits or not package.repo.clean():
            status = []
            if new_commits:
                status.append("new commits")
            if package.repo.modified():
                status.append("modified content")
            if package.repo.untracked():
                status.append("untracked content")
            dirty.append((package, status))
        else:
            clean.append(package)

    for path in ws.root.iterdir():
        if path not in seen_paths and path.is_dir() and GitRepo.is_git_repo(path):
            untracked.append(path)
        seen_paths[path] = True

    log.info("Clean packages:")
    for package in clean:
        log.info("    {}".format(package.name))
    log.info("Dirty packages:")
    for package, content in dirty:
        msg = ", ".join(content)
        log.info("    {} ({})".format(package.name, msg))
    if len(untracked) > 0:
        log.info("Untracked packages:")
        for path in untracked:
            relpath = path.relative_to(ws.root)
            log.info("    {}".format(relpath))
    if len(missing) > 0:
        log.info("Missing packages:")
        for package in missing:
            log.info("    {}".format(package.name))

    packages, errors = ws.resolve()
    for name in packages:
        package = packages[name]
        s = package.status(ws.lock)
        if s:
            print(package.name, s)

    print_errors(errors)


def update(ws, args) -> None:
    packages, errors = ws.resolve(download=True)
    if len(errors) == 0:
        ws.checkout(packages)
    else:
        print_errors(errors)
        sys.exit(1)


def version() -> None:
    version = get_git_version()
    if not version:
        version = get_dist_version()
    print("wit {}".format(version))


def get_git_version():
    # not an official release, use git to get an explicit version
    path = Path(__file__).resolve().parent.parent.parent
    log.spam("Running [git -C {} describe --tags --dirty]".format(str(path)))
    proc = subprocess.run(['git', '-C', str(path), 'describe', '--tags', '--dirty'],
                          stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
    version = proc.stdout.decode('utf-8').rstrip()
    log.spam("Output: [{}]".format(version))
    return re.sub(r"^v", "", version)


def get_dist_version():
    return __version__



#!/usr/bin/env python3

from pathlib import Path
import re
import os
import shutil
from collections import OrderedDict
from .gitrepo import GitRepo, BadSource
from .witlogger import getLogger

log = getLogger()


class WitBug(Exception):
    pass


class Package:
    """ Packages are the "winner" of several dependencies of the same name.
    These "winner" Packages are determined by WorkSpace.resolve and stored in the wit-lock.json
    """

    def __init__(self, name, repo_paths):
        """ Before we know which dependency "won," we need a Package object to
        link these same-named Dependencies together.

        This transitive state is indicated by Package.revision == None

        We need repo_paths here because we will use it to transform sources we need both for
        downloading repos while in the transitive state and for downloading the "winning" git repo
        """

        self.name = name
        self.source = None
        self.revision = None
        self.repo_paths = repo_paths

        self.repo = None
        self.dependents = []

    def set_source(self, source):
        self.source = self.resolve_source(source)

    def short_revision(self):
        if self.revision:
            if self.repo.is_tag(self.revision):
                return self.revision
            return self.repo.get_shortened_rev(self.revision)
        return None

    def __key(self):
        return (self.source, self.revision, self.name)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()

    def add_dependent(self, dep):
        if dep not in self.dependents:
            self.dependents.append(dep)

    def load(self, wsroot, download, source=None, revision=None):
        """Connect a Package to a GitRepo on disk.

        If found, self.repo will be updated.
        """

        source = self.resolve_source(source) or self.resolve_source(self.source)
        revision = revision or self.revision

        assert revision is not None, "Cannot load repo for unknown commit."
        assert source is not None, "Cannot load repo for unknown source."

        # Check if we are already checked out
        self.in_root = (wsroot/self.name).exists()
        if self.in_root:
            repo_root = wsroot
        else:
            repo_root = wsroot/'.wit'
            if not repo_root.exists():
                os.mkdir(str(repo_root))

        self.repo = GitRepo(self.name, repo_root)

        # we carefully use Python's boolean expression evalution short-circuiting
        # to avoid calling has_commit if the repo does not exist
        if (not self.repo.path.exists()
                or not self.repo.has_commit(revision)
                or not (self.repo.is_hash(revision) or self.repo.is_tag(revision))):
            if not download:
                self.repo = None
                return
            try:
                self.repo.download(source, self.name)
            except BadSource:
                self.repo = None
                raise

    def is_ancestor(self, other_commit):
        return self.repo.is_ancestor(other_commit, self.revision)

    def resolve_source(self, source):
        for path in self.repo_paths:
            tmp_path = str(Path(path) / self.name)
            if GitRepo.is_git_repo(tmp_path):
                return tmp_path
        return source

    def get_dependencies(self):
        manifest = self.repo.read_manifest_from_commit(self.revision)
        deps = manifest.dependencies
        for dep in deps:
            dep.add_dependent(self)
        return deps

    def manifest(self):
        res = OrderedDict()
        res['name'] = self.name
        res['source'] = self.source
        res['commit'] = self.revision
        return res

    # this is in Package because update_dependency is in Package
    # it could be confusing to keep the two functions separate
    def add_dependency(self):
        """Change the wit-manifest.json to add a dependency."""
        pass

    def checkout(self, wsroot):
        """Move to root directory and checkout"""
        current_origin = self.repo.get_remote()
        wanted_origin = self.source
        if current_origin != wanted_origin:
            if self.repo.path.parts[-2] == '.wit':
                self.repo.set_origin(self.source)
            else:
                log.warn("Package '{}' wants a different git remote origin.\n"
                         "Origin is currently:\n  {}\n"
                         "Package '{}' wants origin:\n  {}\n"
                         "Please manually update the origin with:\n"
                         "  git -C {} \\\n    remote set-url origin {}"
                         "".format(self.name, current_origin, self.name,
                                   wanted_origin, self.repo.path, wanted_origin))
        assert self.repo.name == self.name
        shutil.move(str(self.repo.path), str(wsroot/self.repo.name))
        self.move_to_root(wsroot)
        self.repo.checkout(self.revision)

    def find_matching_dependent(self):
        """
        Finds first dependent that has the same revision as the resolved package
        """
        if self.revision is None:
            return None
        for dep in self.dependents:
            if dep.specified_revision == self.revision:
                return dep

    def dependents_have_common_ancestor(self):
        commits = [d.specified_revision for d in self.dependents]
        assert self.repo is not None
        return self.repo.have_common_ancestor(commits)

    def move_to_root(self, wsroot: Path):
        assert self.repo.name == self.name
        self.repo.path = wsroot/self.repo.name

    def __repr__(self):
        return "Pkg({})".format(self.id())

    def id(self):
        return "{}::{}".format(self.name, self.short_revision())

    def get_id(self):
        return "pkg_"+re.sub(r"([^\w\d])", "_", self.id())

    def status(self, lock):
        if lock.contains_package(self.name):
            if self.repo and self.revision != self.repo.get_head_commit():
                return "\033[35m(will be checked out to {})\033[m".format(
                    self.short_revision())
        else:
            if not self.in_root:
                return "\033[92m(will be added to workspace and lockfile)\033[m"
            else:
                return "\033[31m(will be added to lockfile)\033[m"



import sys
from .common import print_errors
from .witlogger import getLogger

log = getLogger()


def inspect_tree(ws, args):
    packages, errors = ws.resolve()

    if args.tree:
        tree = {}
        for dep in ws.manifest.dependencies:
            tree[dep.get_id()] = dep.crawl_dep_tree(ws.root, ws.repo_paths, packages)
        for key in tree:
            top_dep = tree[key]
            x, _ = _deduplicate_tree(top_dep)
            _print_generic_tree(x)

    if args.dot:
        _print_dot_tree(ws, packages)

    print_errors(errors)


BOXED_DEPS = False
VERBOSE_GRAPH = False


def _deduplicate_tree(tree, seen=None):
    tree = tree.copy()
    seen = seen or []
    tag = tree.pop('')
    ident = tag[-8:]
    out = {'': tag}

    if ident in seen:
        return out, seen
    else:
        seen.append(ident)
        for key in tree:
            out[key], seen = _deduplicate_tree(tree[key], seen)
        return out, seen


def _print_dot_tree(ws, packages_dict):
    packages = list(packages_dict.values())

    log.output('digraph dependencies {')
    log.output('root [label="[root]"]')

    pkg_ids = []
    for pkg in packages:
        pkg_id = pkg.get_id()
        pkg_ids.append(pkg_id)
        log.output('{} [label="{}"]'.format(pkg_id, pkg.id()))

    drawn_connections = []

    def draw_connection(from_id, to_id, dotted=False):
        if from_id == to_id:
            return
        pair = (from_id, to_id)
        if pair not in drawn_connections:
            log.output("{} -> {}{}".format(from_id, to_id, " [style=dotted]" if dotted else ""))
            drawn_connections.append(pair)

    def print_dep(pkg, dep):
        pkg_id = pkg.get_id()
        dep_id = dep.get_id()
        dep.load(packages_dict, ws.repo_paths, ws.root, False)
        if dep.package.repo is None:
            log.error("Cannot generate graph with missing repo '{}'".format(dep.name))
            sys.exit(1)
        dep_pkg_id = dep.package.get_id()
        if dep.id() != dep.package.id() or VERBOSE_GRAPH:
            draw_connection(dep_id, dep_pkg_id, dotted=True)
            log.output('{} [label="{}"]{}'.format(dep_id, dep.id(),
                                                  " [shape=box]" if BOXED_DEPS else ""))
            draw_connection(pkg_id, dep_id)
        else:
            draw_connection(pkg_id, dep_pkg_id)

    for dep in ws.manifest.dependencies:
        print_dep(ws, dep)

    for pkg in packages:
        for dep in pkg.get_dependencies():
            print_dep(pkg, dep)

    log.output('}')


def _print_generic_tree(data):
    tag = data.pop('')
    print(tag)
    return _recur_print_generic_tree(0, data, [])


def _recur_print_generic_tree(depth, data, done_cols):

    def print_indent(depth):
        for i in range(0, depth):
            if i in done_cols:
                print("   ", end="")
            else:
                print("│  ", end="")

    done_cols_copy = done_cols[:]

    keys = list(data.keys())
    for i, key in enumerate(keys):
        subdata = data[key]
        subtag = subdata.pop('')
        print_indent(depth)
        if i == len(keys)-1:
            print("└─", end="")
            done_cols_copy.append(depth)
        else:
            print("├─", end="")
        print(subtag)
        _recur_print_generic_tree(depth+1, subdata, done_cols_copy)



#!/usr/bin/env python3

import logging
import sys
from typing import cast

VERBOSE = 15
TRACE = 5
SPAM = 3


def getLogger():
    """
    Get the Wit logger
    """
    # We have to cast to make mypy happy
    return cast(WitLogger, logging.getLogger('wit'))


class WitFormatter(logging.Formatter):
    """
    Custom formatter to prefix messages except for log.info
    """

    info_format = "%(msg)s"

    def __init__(self):
        super().__init__(fmt="[%(levelname)s] %(msg)s", datefmt=None, style='%')

    def format(self, record):

        orig_fmt = self._style._fmt
        if record.levelno == logging.INFO:
            self._style._fmt = WitFormatter.info_format

        result = logging.Formatter.format(self, record)
        self._style._fmt = orig_fmt

        return result


class WitLogger(logging.Logger):
    # See: https://stackoverflow.com/a/22586200/1785651

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
        # Use WitFormatter for info level
        _handler = logging.StreamHandler(sys.stdout)
        _handler.setFormatter(WitFormatter())
        logging.basicConfig(level=logging.INFO, handlers=[_handler])

        logging.addLevelName(VERBOSE, 'VERBOSE')
        logging.addLevelName(TRACE, 'TRACE')
        logging.addLevelName(SPAM, 'SPAM')

    def getLevelName(self):
        return logging.getLevelName(getLogger().getEffectiveLevel())

    def verbose(self, msg, *args, **kwargs):
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, msg, args, **kwargs)

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)

    def spam(self, msg, *args, **kwargs):
        if self.isEnabledFor(SPAM):
            self._log(SPAM, msg, args, **kwargs)

    # for potential future -o flags (e.g. for `wit inspect --dot`)
    def output(self, msg):
        print(msg)


logging.setLoggerClass(WitLogger)



#!/usr/bin/env python3

import json
from .gitrepo import GitRepo
from collections import OrderedDict
from typing import Optional
from .witlogger import getLogger
from .package import Package

log = getLogger()


# TODO
# Should we use different datastructures?
# The JSON file format slightly differs from manifest, why?
class LockFile:
    """
    Common class for the description of package dependencies and a workspace
    """

    def __init__(self, packages=[]):
        self.packages = packages

    def get_package(self, name: str) -> Optional[GitRepo]:
        for p in self.packages:
            if p.name == name:
                return p
        return None

    def contains_package(self, name: str) -> bool:
        return self.get_package(name) is not None

    def add_package(self, package):
        self.packages.append(package)

    def write(self, path):
        log.debug("Writing lock file to {}".format(path))
        contents = OrderedDict((p.name, p.manifest()) for p in self.packages)
        manifest_json = json.dumps(contents, sort_keys=True, indent=4) + '\n'
        path.write_text(manifest_json)

    @staticmethod
    def read(path):
        log.debug("Reading lock file from {}".format(path))
        content = json.loads(path.read_text())
        return LockFile.process(content)

    @staticmethod
    def process(content):
        pkgs = [lockfile_item_to_pkg(x) for _, x in content.items()]
        return LockFile(pkgs)


if __name__ == '__main__':
    import doctest
    doctest.testmod()


def lockfile_item_to_pkg(item):
    pkg = Package(item['name'], [])
    pkg.set_source(item['source'])
    pkg.revision = item['commit']
    return pkg



import re
import multiprocessing.dummy
from datetime import datetime
from pathlib import Path
from typing import List  # noqa: F401
from collections import OrderedDict
from .common import WitUserError
from .package import Package
from .witlogger import getLogger
from .gitrepo import BadSource

log = getLogger()


class DependeeNewerThanDepender(WitUserError):
    def __init__(self, depender, dependee):
        self.depender = depender
        self.dependee = dependee

    def __str__(self):
        return ("Depender {} is older than its dependee {}\n"
                "This should not happen, but it may be caused by a fictitious "
                "clock time being stored in a commit.\n This should be fixable "
                "by creating a new commit in the dependee then depending on "
                "that commit."
                "".format(self.depender.id(), self.dependee.id()))


class Dependency:
    """ A dependency that a Package specifies. From wit-manifest.json and wit-workspace.js """

    def __init__(self, name, source, specified_revision, message):
        self.source = source
        self.specified_revision = specified_revision or "HEAD"
        self.name = name or Dependency.infer_name(source)
        self.package = None  # type: Package
        self.dependents = []  # type: List[Package]
        self.message = message

    def resolve_deps(self, wsroot, repo_paths, download, source_map, packages, queue, jobs):
        source_map = source_map.copy()
        packages = packages.copy()
        queue = queue.copy()
        subdeps = self.package.get_dependencies()
        log.debug("Dependencies for [{}]: [{}]".format(self.name, subdeps))

        errors = self._parallel_clone(subdeps, wsroot, repo_paths, download, jobs)

        for subdep in subdeps:
            subdep.load(packages, repo_paths, wsroot, download=False)
            sources_conflict_check(subdep, source_map)
            source_map[subdep.name] = subdep.package.resolve_source(subdep.source)

            if subdep.package.repo is None:
                continue

            if subdep.get_commit_time() > self.get_commit_time():
                errors.append(DependeeNewerThanDepender(self, subdep))
                continue

            commit_time = subdep.get_commit_time()
            queue.append((commit_time, subdep))

        queue.sort(key=lambda tup: tup[0])

        return source_map, packages, queue, errors

    def _parallel_clone(self, deps, wsroot, repo_paths, download, jobs):
        errors = []

        def do(dep):
            try:
                p = Package(dep.name, repo_paths)
                p.load(wsroot, download, dep.source, dep.specified_revision)
            except BadSource as e:
                errors.append(e)

        with multiprocessing.dummy.Pool(jobs) as pool:
            pool.map(do, deps)

        return errors

    def __key(self):
        return (self.source, self.specified_revision, self.name)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(self, type(other)) and self.__key() == other.__key()

    @staticmethod
    def infer_name(source):
        return Path(source).name.replace('.git', '')

    # NB: mutates packages[self.name]
    def load(self, packages, repo_paths, wsroot, download):
        if self.name in packages:
            self.package = packages[self.name]
        else:
            self.package = Package(self.name, repo_paths)
            packages[self.name] = self.package
        self.package.add_dependent(self)

        self.package.load(wsroot, download, source=self.source, revision=self.specified_revision)

    def add_dependent(self, dependent):
        if dependent not in self.dependents:
            self.dependents.append(dependent)

    def get_commit_time(self):
        return datetime.utcfromtimestamp(int(self.package.repo.commit_to_time(
            self.specified_revision)))

    def manifest(self):
        res = OrderedDict()
        res['name'] = self.name
        res['source'] = self.source
        res['commit'] = self.specified_revision
        if self.message is not None and len(self.message) > 0:
            res['//'] = self.message
        return res

    # used before saving to manifests/lockfiles
    def resolved(self):
        return Dependency(self.name, self.source, self.resolved_rev(), self.message)

    # Check if the Dependency has a Package and repo on disk
    def _is_bound(self) -> bool:
        return self.package is not None and self.package.repo is not None

    def resolved_rev(self):
        if not self._is_bound():
            raise Exception("Cannot resolve dependency that is unbound to disk")
        return self.package.repo.get_commit(self.specified_revision)

    def __repr__(self):
        return "Dep({})".format(self.id())

    def short_revision(self):
        if self.package and self.package.repo:
            if self.package.repo.is_hash(self.specified_revision):
                return self.package.repo.get_shortened_rev(self.specified_revision)
            return self.specified_revision
        return self.specified_revision[:8]

    def id(self):
        return "{}::{}".format(self.name, self.short_revision())

    def get_id(self):
        return "dep_"+re.sub(r"([^\w\d])", "_", self.id())

    def crawl_dep_tree(self, wsroot, repo_paths, packages):
        fancy_tag = self.id()
        self.load(packages, repo_paths, wsroot, False)
        if self.package.repo is None:
            return {'': "{} \033[91m(missing)\033[m".format(fancy_tag)}
        if self.package.revision != self.resolved_rev():
            fancy_tag += "->{}".format(self.package.short_revision())
            return {'': fancy_tag}

        tree = {'': fancy_tag}
        subdeps = self.package.get_dependencies()
        for subdep in subdeps:
            tree[subdep.get_id()] = subdep.crawl_dep_tree(wsroot, repo_paths, packages)
        return tree


def sources_conflict_check(dep, source_map):
    if dep.name in source_map:
        dep_resolved_source = dep.package.resolve_source(dep.source)
        if dep_resolved_source != source_map[dep.name]:
            if not dep.package.dependents_have_common_ancestor():
                raise WitUserError(("Two dependencies have the same name "
                                    "but an unrelated git history:\n"
                                    "  {}\n"
                                    "  {}\n"
                                    "".format(dep_resolved_source,
                                              source_map[dep.name])))


def parse_dependency_tag(s):
    parts = s.split('::')
    source = parts[0]
    rev = parts[1] if parts[1:] else None

    return source, rev


def manifest_item_to_dep(obj):
    # source can be done due to repo path
    return Dependency(obj['name'], obj.get('source', None), obj['commit'], obj.get('//', None))



import argparse
import os
from .dependency import parse_dependency_tag

# default max parallel git clones possible by 'init' or 'update'
_max_clone_jobs = 64


def chdir(s) -> None:
    def err(msg):
        raise argparse.ArgumentTypeError(msg)
    try:
        os.chdir(s)
    except FileNotFoundError:
        err("'{}' path not found!".format(s))
    except NotADirectoryError:
        err("'{}' is not a directory!".format(s))


# ********** top-level parser **********
parser = argparse.ArgumentParser(
    prog='wit',
    formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument('-v', '--verbose', action='count', default=0,
                    help='''Specify level of verbosity
-v:    verbose
-vv:   debug
-vvv:  trace
-vvvv: spam
''')
parser.add_argument('--version', action='store_true', help='Print wit version')
parser.add_argument('-C', dest='cwd', type=chdir, metavar='path', help='Run in given path')
parser.add_argument('--repo-path', default=os.environ.get('WIT_REPO_PATH'),
                    help='Specify alternative paths to look for packages')
parser.add_argument('--prepend-repo-path', default=None,
                    help='Prepend paths to the default repo search path.')
parser.add_argument('-j', '--max-parallel-clones', dest='jobs', default=_max_clone_jobs, type=int,
                    help="Max quantity of 'git clone' to run in parallel. "
                    "Default is '{}'. Set to '1' for serial cloning.".format(_max_clone_jobs))

# ********** command subparser aggregator **********
subparsers = parser.add_subparsers(
               title='subcommands',
               dest='command',
               metavar='<command>',
               help='<description>')

# ********** init subparser **********
init_parser = subparsers.add_parser('init', help='create workspace')
init_parser.add_argument('--no-update', dest='update', action='store_false',
                         help='don\'t run update upon creating the workspace')
init_parser.add_argument('-a', '--add-pkg', metavar='repo[::revision]', action='append',
                         type=parse_dependency_tag, help='add an initial package')
init_parser.add_argument('workspace_name')

# ********** restore subparser **********
restore_parser = subparsers.add_parser(
    'restore',
    help='restore previous workspace',
    description='Create a workspace in the current directory from the metadata of a previous '
                'workspace. The wit-workspace.json and wit-lock.json files are required to be '
                'provided but are assumed to be in the current directory.')
restore_parser.add_argument(
    '-n', '--name', dest='workspace_name',
    help='new directory to create workspace in. Default is $PWD.')
restore_parser.add_argument('-w', '--from-workspace', dest='from_workspace',
                            help="directory containing wit-lock.json and wit-workspace.json")

# ********** add-pkg subparser **********
add_pkg_parser = subparsers.add_parser('add-pkg', help='add a package to the workspace')
add_pkg_parser.add_argument('repo', metavar='repo[::revision]', type=parse_dependency_tag)

# ********** update-pkg subparser **********
update_pkg_parser = subparsers.add_parser('update-pkg', help='update the revision of a '
                                          'previously added package')
update_pkg_parser.add_argument('repo', metavar='repo[::revision]', type=parse_dependency_tag)

# ********** add-dep subparser **********
add_dep_parser = subparsers.add_parser(
    name='add-dep',
    description='Adds <pkg> as a dependency to a target package determined by the current working '
                'directory (which can be set by -C)',
    help='add a dependency to a package')
add_dep_parser.add_argument(
    dest='pkg',
    metavar='pkg[::revision]',
    type=parse_dependency_tag,
    help='revision can be any git commit-ish, default is the currently checked out commit')
add_dep_parser.add_argument(
    '-m',
    '--message',
    dest="message",
    help="Comment message to be added to the dependency's entry in the manifest.")

# ********** update-dep subparser **********
update_dep_parser = subparsers.add_parser('update-dep', help='update revision of a dependency '
                                          'in a package')
update_dep_parser.add_argument('pkg', metavar='pkg[::revision]', type=parse_dependency_tag)
update_dep_parser.add_argument(
    '-m',
    '--message',
    dest="message",
    help='Comment message to be added to the dependency\'s entry in the manifest. This will'
         ' overwrite a previous message.')

# ********** status subparser **********
subparsers.add_parser('status', help='show status of workspace')

# ********** update subparser **********
subparsers.add_parser('update', help='update git repos')

# ********** inspect subparser **********
inspect_parser = subparsers.add_parser('inspect', help='inspect lockfile')
inspect_group = inspect_parser.add_mutually_exclusive_group()
inspect_group.add_argument('--tree', action="store_true")
inspect_group.add_argument('--dot', action="store_true")

# ********** foreach subparser **********
foreach_parser = subparsers.add_parser(
    'foreach',
    help='perform a command in each repository directory',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""
Perform a command in each repository directory.
The repository list is created by reading records contained in 'wit-lock.json'.

Any options, such as --continue-on-fail must be specified before the command.

Wit sets the following environment variables for each invocation of the command:
    WIT_REPO_NAME    repository name
    WIT_REPO_PATH    path to repository
    WIT_LOCK_SOURCE  initial source location of repository
    WIT_LOCK_COMMIT  commit recorded in lockfile, actual repository contents could be different
    WIT_WORKSPACE    path to root of wit workspace""")

foreach_parser.add_argument('--continue-on-fail', action='store_true',
                            help='run the command in each repository regardless of failures')

# 'cmd' and 'args' eventually become one list, but this forces at least one input string
foreach_parser.add_argument('cmd', help='command to run in each repository')
foreach_parser.add_argument('args', nargs=argparse.REMAINDER, help='arguments for the command')



#!/usr/bin/env python3

import sys
from .witlogger import getLogger

log = getLogger()


def error(*args, **kwargs):
    log.error(*args, **kwargs)
    sys.exit(1)


def print_errors(errors):
    if len(errors) > 0:
        log.info("")  # print newline

    for err in errors:
        log.info("--- ERROR ---")
        log.info(err)


class WitUserError(Exception):
    """
    Supertype of user-input errors that should be reported without stack traces
    """
    pass



import setuptools

version = {}
with open("wit/version.py") as fd:
    exec(fd.read(), version)

setuptools.setup(
    name='wit-sifive',
    version=version['__version__'],
    description="Wit: Workspace Integration Tool",
    long_description="See README.md at https://github.com/sifive/wit/",
    author='SiFive',
    author_email='',
    url='https://www.github.com/sifive/wit',
    packages=["wit"],
    entry_points={'console_scripts': ['wit=wit.main:main']},
    python_requires='>=3.5',
    classifiers=["License :: OSI Approved :: Apache Software License"],
)



#!/usr/bin/env python3

import sys

if sys.version_info[0] < 3 or sys.version_info[1] < 5:
    print("This program requires Python version 3.5 or newer")
    sys.exit(1)

if __name__ == '__main__':
    from .main import main
    main()



#!/usr/bin/env python3

import sys
import shutil
import threading
import queue
from pathlib import Path
from pprint import pformat
from .manifest import Manifest
from .dependency import Dependency, sources_conflict_check
from .lock import LockFile
from .common import WitUserError, error
from .witlogger import getLogger
from .gitrepo import GitCommitNotFound

log = getLogger()


# TODO: hide the stacktrace when not debugging
class NotAncestorError(WitUserError):
    def __init__(self, orig_child: Dependency, old_child: Dependency):
        self.orig_child = orig_child
        self.old_child = old_child

    def __str__(self):
        # orig is later in time because we traverse the queue backwards in time
        assert self.orig_child.name == self.old_child.name
        child_name = self.old_child.name

        orig_parent = self.orig_child.dependents[0].dependents[0]
        old_parent = self.old_child.dependents[0].dependents[0]
        return ("\n\nAncestry error:\n"
                "'{orig_parent_name}' and '{old_parent_name}' both depend on '{child_name}':\n"
                "    {orig_parent_tag} depends on "
                "{orig_child_tag}\n"
                "    {old_parent_tag} depends on "
                "{old_child_tag}\n\n"
                "Although {orig_child_tag} is newer than "
                "{old_child_tag},\n{orig_child_tag} is not "
                "a descendent of {old_child_tag}.\n\n"
                "Therefore, there is no guarantee that "
                "the dependee needed by {old_parent_tag} will be satisfied "
                "by the dependee needed by {orig_parent_tag}."
                "".format(
                    orig_parent_name=orig_parent.name,
                    old_parent_name=old_parent.name,

                    orig_parent_tag=orig_parent.id(),
                    old_parent_tag=old_parent.id(),

                    child_name=child_name,
                    orig_child_tag=self.orig_child.id(),
                    old_child_tag=self.old_child.id(),
                ))


class PackageNotInWorkspaceError(WitUserError):
    pass


class WorkSpace:
    MANIFEST = "wit-workspace.json"
    LOCK = "wit-lock.json"

    def __init__(self, root, repo_paths, jobs=None):
        self.root = root
        self.repo_paths = repo_paths
        self.manifest = self._load_manifest()
        self.lock = self._load_lockfile()
        self.jobs = jobs

    def id(self):
        return "[root]"

    def get_id(self):
        return "root"

    @classmethod
    def create(cls, name, repo_paths, jobs):
        """Create a wit workspace on disk with the appropriate json files"""
        root = Path.cwd() / name
        manifest_path = cls._manifest_path(root)
        if root.exists():
            log.info("Using existing directory [{}]".format(str(root)))

            if manifest_path.exists():
                log.error("Manifest file [{}] already exists.".format(manifest_path))
                sys.exit(1)
        else:
            log.info("Creating new workspace [{}]".format(str(root)))
            try:
                root.mkdir()
            except Exception as e:
                log.error("Unable to create workspace [{}]: {}".format(str(root), e))
                sys.exit(1)

        dotwit = root/'.wit'
        if dotwit.exists():
            # we could keep the old cached repos, but if the user is explicitly re-initing,
            # they probably want a 100% clean slate
            shutil.rmtree(str(dotwit))
        dotwit.mkdir()

        manifest = Manifest([])
        manifest.write(manifest_path)

        lockfile = LockFile([])
        lockfile.write(cls._lockfile_path(root))

        return WorkSpace(root, repo_paths, jobs)

    @classmethod
    def restore(cls, root):
        # constructing WorkSpace will parse the lock file
        ws = WorkSpace(root, [])

        def do_clone(pkg, root, errors):
            try:
                pkg.load(root, True)
                pkg.checkout(root)
            except Exception as e:
                errors.put(e)

        errors = queue.Queue()
        threads = list()
        for pkg in ws.lock.packages:
            t = threading.Thread(target=do_clone, args=(pkg, root, errors))
            threads.append(t)
            t.start()

        for t in threads:
            t.join()

        if not errors.empty():
            while not errors.empty():
                e = errors.get()
                log.error("Unable to create workspace [{}]: {}".format(str(root), e))
            sys.exit(1)

        return ws

    def _load_manifest(self):
        return Manifest.read_manifest(self.manifest_path())

    def _load_lockfile(self):
        return LockFile.read(self.lockfile_path())

    @classmethod
    def _manifest_path(cls, root):
        return root / cls.MANIFEST

    def manifest_path(self):
        return WorkSpace._manifest_path(self.root)

    @classmethod
    def _lockfile_path(cls, path):
        return path / cls.LOCK

    def lockfile_path(self):
        return WorkSpace._lockfile_path(self.root)

    @staticmethod
    def find(start, repo_paths, jobs):
        cwd = start.resolve()
        for p in ([cwd] + list(cwd.parents)):
            manifest_path = WorkSpace._manifest_path(p)
            log.debug("Checking [{}]".format(manifest_path))
            if Path(manifest_path).is_file():
                log.debug("Found workspace at [{}]".format(p))
                return WorkSpace(p, repo_paths, jobs)

        raise FileNotFoundError("Couldn't find workspace file")

    def resolve(self, download=False):
        source_map, packages, queue = \
            self.resolve_deps(self.root, self.repo_paths, download, {}, {}, [])

        errors = list()
        dep_errors = list()
        while queue:
            commit_time, dep = queue.pop()
            log.debug("{} {}".format(commit_time, dep))

            name = dep.package.name
            if name in packages and packages[name].revision is not None:
                package = packages[name]
                if not package.repo.is_ancestor(dep.specified_revision, package.revision):
                    errors.append(NotAncestorError(package.find_matching_dependent(), dep))
                continue

            packages[dep.name] = dep.package
            packages[dep.name].revision = dep.resolved_rev()
            packages[dep.name].set_source(dep.source)

            source_map, packages, queue, dep_errors = \
                dep.resolve_deps(self.root, self.repo_paths, download, source_map,
                                 packages, queue, self.jobs)

        for pkg in packages.values():
            if not pkg.repo or pkg.repo.path.parts[-2] == '.wit':
                continue

            used_commit = pkg.revision
            fs_commit = pkg.repo.get_commit('HEAD')
            if used_commit != fs_commit:
                log.warn("using '{}' manifest instead of checked-out version of '{}'".format(
                    pkg.id(), pkg.name))
                continue

            if pkg.repo.modified_manifest():
                log.warn("disregarding uncommitted changes to the '{}' manifest".format(pkg.name))

        return packages, errors + dep_errors

    def resolve_deps(self, wsroot, repo_paths, download, source_map, packages, queue):
        source_map = source_map.copy()
        queue = queue.copy()
        for dep in self.manifest.dependencies:
            dep.load(packages, repo_paths, wsroot, download)

            sources_conflict_check(dep, source_map)

            source_map[dep.name] = dep.source

            commit_time = dep.get_commit_time()
            queue.append((commit_time, dep))

        queue.sort(key=lambda tup: tup[0])

        return source_map, packages, queue

    def checkout(self, packages):
        lock_packages = []
        for name in packages:
            package = packages[name]
            package.checkout(self.root)
            lock_packages.append(package)

        new_lock = LockFile(lock_packages)
        new_lock_path = WorkSpace._lockfile_path(self.root)
        new_lock.write(new_lock_path)
        self.lock = new_lock

    def add_dependency(self, tag) -> None:
        """ Resolve a dependency then add it to the wit-workspace.json """
        from .main import dependency_from_tag
        dep = dependency_from_tag(self.root, tag)

        if self.manifest.contains_dependency(dep.name):
            error("Manifest already contains package {}".format(dep.name))

        packages = {pkg.name: pkg for pkg in self.lock.packages}
        dep.load(packages, self.repo_paths, self.root, True)
        try:
            dep.package.revision = dep.resolved_rev()
        except GitCommitNotFound:
            raise WitUserError("Could not find commit or reference '{}' in '{}'"
                               "".format(dep.specified_revision, dep.name))

        assert dep.package.repo is not None

        self.manifest.add_dependency(dep)

        log.debug('my manifest_path = {}'.format(self.manifest_path()))
        self.manifest.write(self.manifest_path())

        log.info("The workspace now depends on '{}'".format(dep.package.id()))

    def update_dependency(self, tag) -> None:
        # init requested Dependency
        from .main import dependency_from_tag
        req_dep = dependency_from_tag(self.root, tag)

        manifest_dep = self.manifest.get_dependency(req_dep.name)

        # check if the package is missing from the wit-workspace.json
        if manifest_dep is None:
            log.error("Package {} not in wit-workspace.json".format(req_dep.name))
            log.error("Did you mean to run 'wit add-pkg' or 'wit update-dep'?")
            sys.exit(1)

        # load their Package
        packages = {pkg.name: pkg for pkg in self.lock.packages}
        req_dep.load(packages, self.repo_paths, self.root, True)

        manifest_dep.load(packages, self.repo_paths, self.root, True)

        # check if the dependency is missing from disk
        if req_dep.package.repo is None:
            msg = "Cannot update package '{}'".format(req_dep.name)
            if self.lock.contains_package(req_dep.name):
                msg += (":\nAlthough '{}' exists (according to the wit-lock.json), "
                        "it has not been cloned to the root workspace.").format(req_dep.name)
            else:
                msg += "because it does not exist in the workspace."
            raise PackageNotInWorkspaceError(msg)

        try:
            req_dep.package.revision = req_dep.resolved_rev()
        except GitCommitNotFound:
            raise WitUserError("Could not find commit or reference '{}' in '{}'"
                               "".format(req_dep.specified_revision, req_dep.name))

        # compare the requested revision to the revision in the wit-workspace.json
        if manifest_dep.resolved_rev() == req_dep.package.revision:
            log.warn("Updating '{}' to the same revision it already is!".format(req_dep.name))

        self.manifest.replace_dependency(req_dep)
        self.manifest.write(self.manifest_path())

        log.info("The workspace now depends on '{}'".format(req_dep.package.id()))

        # if we differ from the lockfile, tell the user to update
        if ((not self.lock.contains_package(req_dep.name) or
             not self.lock.get_package(req_dep.name).revision == req_dep.package.revision)):
            log.info("Don't forget to run 'wit update'!")

    # Enable prettyish-printing of the class
    def __repr__(self):
        return pformat(vars(self), indent=4, width=1)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
